// ************************
// * Half-Life IRC Plugin *
// ************************
// version 4.0
// by Pierre-Marie Baty <pm@bots-united.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

// TCP socket headers and definitions
#ifdef _WIN32
#include <winsock2.h>
#else
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#define closesocket close
#endif

// multithreading headers and definitions
#ifdef _WIN32
#include <process.h>
#ifndef _MT
#define _MT
#endif
#define MT_void void
#else
#include <pthread.h>
#define MT_void void *
#endif

// sleep routines
#ifdef _WIN32
#ifndef sleep
#define sleep(x) Sleep (x * 1000)
#endif
#endif


// IRC bot orders
#define IRCBOT_ORDER_NOORDER 0
#define IRCBOT_ORDER_CONNECT 1
#define IRCBOT_ORDER_SETNICKNAME 2
#define IRCBOT_ORDER_JOINCHANNEL 3
#define IRCBOT_ORDER_DISCONNECT 4
#define IRCBOT_ORDER_TERMINATETHREAD 5

// IRC bot errors
#define IRCBOT_ERROR_NOERROR 0
#define IRCBOT_ERROR_CANTRESOLVENAME 1
#define IRCBOT_ERROR_CANTCREATESOCKET 2
#define IRCBOT_ERROR_CANTCONNECTTOSERVER 3


typedef struct
{
   char hostname[256]; // IRC server host name
   int port; // IRC server port
   sockaddr_in sockaddr; // IRC server socket address
   hostent *hostinfo; // IRC server host info
} ircserver_t;


typedef struct
{
   int order; // current order given to the IRC bot
   int error; // current error reported by the IRC bot
   bool thread_active; // set to TRUE if the IRC bot thread is currently running
   bool connected_to_server; // set to TRUE if the IRC bot is connected to the IRC server
   bool connected_to_channel; // set to TRUE if the IRC bot is currently member of the IRC channel
   float connection_idle_delay; // increasing sleep delay between disconnections and reconnections
   char channel[256]; // name of the IRC channel this bot is supposed to join
   char nickname[32]; // IRC nickname of the IRC bot in all characters
   char irc_to_game[1024]; // chat string from an IRC member to be displayed on the players' HUDs
   bool new_irc_to_game; // set to TRUE when the previous string holds new (unprocessed) content
   char game_to_irc[1024]; // chat string from a player to be sent to the IRC channel
   bool new_game_to_irc; // set to TRUE when the previous string holds new (unprocessed) content
   char tx_buffer[1024]; // chat string to be sent to the IRC channel by the IRC bot
   char rx_buffer[1024]; // chat string received from the IRC channel by the IRC bot
   unsigned int irc_socket; // network socket ID of the IRC bot client
#ifdef _WIN32
   WSADATA wsaData; // WinSock data setup for the IRC bot client socket
#endif
} ircbot_t;


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
ircserver_t ircserver;
ircbot_t ircbot;
bool playerdeath_sent[32];
#ifndef _WIN32
pthread_t ircbot_thread;
#endif

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void ServerActivate (edict_t *pEdictList, int edictCount, int clientMax);
void ClientDisconnect (edict_t *pEntity);
void ClientPutInServer (edict_t *pEntity);
void ClientCommand (edict_t *pEntity);
void StartFrame (void);
void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed);
void WriteString (const char *sz);
void ServerCommand_IRC (void);
void SendIRCToGame (const char *text);
void SendGameToIRC (const char *text);
void ServerConsole_printf (char *fmt, ...);
// multithreaded functions
MT_void StartMultithreadedIRCBot (void *thread_id);
void IRCBot_Connect (void);
void IRCBot_Disconnect (void);
void IRCBot_SetNickname (void);
void IRCBot_JoinChannel (void);
void IRCBot_Ping (void);
void IRCBot_Talk (void);
void IRCBot_Listen (void);
void IRCBot_Think (void);
const char *GetWord (const char *string, int word_number);
int GetNextWordStartOffset (const char *string, int word_number);


// START of Metamod stuff

gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   NULL, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "HLIRC", // plugin name
   "5.0", // plugin version
   "02/07/2004", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "HLIRC", // plugin logtag
   PT_ANYTIME, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   int dummy = 0;

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

#ifdef _WIN32
   // initialize Windows TCP stack (Winsock)
   if (WSAStartup (0x0101, &ircbot.wsaData) != 0)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (plugin failed to initialize Winsock)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (plugin failed to initialize Winsock)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }
#endif

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // reset the IRC server structure with some default values
   memset (&ircserver, 0, sizeof (ircserver));
   strcpy (ircserver.hostname, "localhost"); // set a default server hostname
   ircserver.port = 6667; // set a default server port

   // reset the IRC bot structure, launch the thread and wait for it to start
   memset (&ircbot, 0, sizeof (ircbot));
   strcpy (ircbot.nickname, "HLServer"); // set a default nickname
   strcpy (ircbot.channel, "#hlserver"); // set a default channel
   ircbot.connection_idle_delay = 5.0; // IRC bot will connect in 5 seconds only

   // launch the IRC bot thread (Win32 thread API or POSIX threads)
#ifdef _WIN32
   _beginthread (StartMultithreadedIRCBot, 0, NULL);
#else
   pthread_create (&ircbot_thread, NULL, StartMultithreadedIRCBot, (void *) dummy);
#endif

   while (!ircbot.thread_active)
      sleep (0); // wait for thread to start

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the server commands this plugin uses
   REG_SVR_COMMAND ("irc", ServerCommand_IRC);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

#ifdef _WIN32
   WSACleanup (); // unallocate the Winsock ressources allocated by WSAStartup()
#endif

   // terminate the IRC bot thread
   ircbot.order = IRCBOT_ORDER_TERMINATETHREAD; // order the IRC bot to suicide
   IRCBot_Ping (); // make the bot send a ping so it wakes up when it receives the bounce

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   // this is one of the initialization functions hooked by metamod in the gamedll API

   // keep track of pointers to the functions we use in the gamedll API
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnClientPutInServer = ClientPutInServer;
   gFunctionTable.pfnClientCommand = ClientCommand;
   gFunctionTable.pfnServerActivate = ServerActivate;
   gFunctionTable.pfnStartFrame = StartFrame;

   // copy the whole table for metamod to know which functions we are using here
   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE); // alright
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   // required Win32 DLL entry point. Good place for freeing stuff upon DLL unloading (in case
   // fdwReason would be DLL_PROCESS_DETACH) if needed. The same for Linux would be implemented
   // in an optional _fini() function.

   return (TRUE); // nothing to do here
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   // this function is called by the engine to tell the MOD DLL (hence the plugins too) the
   // address of its function table and its server global variables structure.

   // keep track of the engine functions and the server globals for our own use
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


void ClientDisconnect (edict_t *pEntity)
{
   // this function is called whenever a client is VOLUNTARILY disconnected from the server,
   // either because the client dropped the connection, or because the server dropped him from
   // the game (latency timeout). The effect is the freeing of a client slot on the server.

   sprintf (ircbot.game_to_irc, "*** %s has left the game ***\n", STRING (pEntity->v.netname));
   ircbot.new_game_to_irc = TRUE; // we have something new to send

   RETURN_META (MRES_IGNORED);
}


void ClientPutInServer (edict_t *pEntity)
{
   // this function is called once a just connected client actually enters the server, after
   // having downloaded and synchronized its resources with the of the server's. It's the
   // perfect place to hook for client connecting, since a client can always try to connect
   // passing the ClientConnect() step, and not be allowed by the server later (because of a
   // latency timeout or whatever reason).

   sprintf (ircbot.game_to_irc, "*** %s joined the game ***\n", STRING (pEntity->v.netname));
   ircbot.new_game_to_irc = TRUE; // we have something new to send

   RETURN_META (MRES_IGNORED);
}


void ClientCommand (edict_t *pEntity)
{
   // this function is called whenever a Half-Life client sends a client command to the server.
   // The command arguments can be fetched using the CMD_ARGV and CMD_ARGC facilities.

   char cmd[128];
   char arg1[128];

   strcpy (cmd, CMD_ARGV (0)); // I hate pointers in these situations
   strcpy (arg1, CMD_ARGV (1));

   // is this a "say" command ?
   if ((strcmp (cmd, "say") == 0) && (arg1[0] != 0))
   {
      sprintf (ircbot.game_to_irc, "%s%c%s%c: %s",
               (pEntity->v.deadflag != DEAD_NO ? "(DEAD) " : ""),
               0x02, STRING (pEntity->v.netname), 0x02,
               arg1);
      ircbot.new_game_to_irc = TRUE; // tell the IRC bot it has a new message to send
   }

   // else is this a "say_team" command ?
   else if ((strcmp (cmd, "say_team") == 0) && (arg1[0] != 0))
   {
      sprintf (ircbot.game_to_irc, "%s%c%s%c (team only): %s",
               (pEntity->v.deadflag != DEAD_NO ? "(DEAD) " : ""),
               0x02, STRING (pEntity->v.netname), 0x02,
               arg1);
      ircbot.new_game_to_irc = TRUE; // tell the IRC bot it has a new message to send
   }

   RETURN_META (MRES_IGNORED);
}


void ServerActivate (edict_t *pEdictList, int edictCount, int clientMax)
{
   // this function is called when the server has fully loaded and is about to manifest itself
   // on the network as such. Since a mapchange is actually a server shutdown followed by a
   // restart, this function is also called when a new map is being loaded. Once this function
   // has been called, the server can be considered as "running".

   sprintf (ircbot.game_to_irc, "*** booting new map: %s ***\n", STRING (gpGlobals->mapname));
   ircbot.new_game_to_irc = TRUE; // we have something new to send

   RETURN_META (MRES_IGNORED);
}


void StartFrame (void)
{
   // this function starts a video frame. It is called once per video frame by the engine. If
   // you run Half-Life at 90 fps, this function will then be called 90 times per second. By
   // placing a hook on it, we have a good place to do things that should be done continuously
   // during the game, for instance keeping an eye on the IRC bot, and checking whether a player
   // was killed by someone else.

   int player_index;
   edict_t *pPlayer;
   static char weapon_name[64];
   int index, length;

   // is the IRC bot complaining about something ?
   if (ircbot.error == IRCBOT_ERROR_CANTRESOLVENAME)
   {
      LOG_CONSOLE (PLID, "IRC bot can't resolve IRC server name"); // daaaaah
      LOG_ERROR (PLID, "IRC bot can't resolve IRC server name"); // mommyyy
      ircbot.error = IRCBOT_ERROR_NOERROR;
   }
   else if (ircbot.error == IRCBOT_ERROR_CANTCREATESOCKET)
   {
      LOG_CONSOLE (PLID, "IRC bot can't create socket"); // daaaaah
      LOG_ERROR (PLID, "IRC bot can't create socket"); // mommyyy
      ircbot.error = IRCBOT_ERROR_NOERROR;
   }
   else if (ircbot.error == IRCBOT_ERROR_CANTCONNECTTOSERVER)
   {
      LOG_CONSOLE (PLID, "IRC bot can't connect to server"); // daaaaah
      LOG_ERROR (PLID, "IRC bot can't connect to server"); // mommyyy
      ircbot.error = IRCBOT_ERROR_NOERROR;
   }

   // cycle through all players
   for (player_index = 0; player_index < gpGlobals->maxClients; player_index++)
   {
      pPlayer = INDEXENT (player_index + 1); // quick access to player

      if (FNullEnt (pPlayer))
         continue; // skip invalid player slots

      // was this player killed (normal death or gibbed) ?
      if ((pPlayer->v.deadflag == DEAD_DYING) || (pPlayer->v.health < -1))
      {
         // have we NOT notified the IRC channel yet ?
         if (!playerdeath_sent[player_index])
         {
            // has the player killed self ?
            if (FNullEnt (pPlayer->v.dmg_inflictor)
                || (pPlayer->v.dmg_inflictor == pPlayer)
                || (strcmp ("trigger_hurt", STRING (pPlayer->v.dmg_inflictor->v.classname)) == 0)
                || (strcmp ("worldspawn", STRING (pPlayer->v.dmg_inflictor->v.classname)) == 0))
            {
               sprintf (ircbot.game_to_irc,
                        "*** %s killed self ! ***\n",
                        STRING (pPlayer->v.netname));
               ircbot.new_game_to_irc = TRUE; // we have something new to send
            }

            // else is the killer a player ?
            else if (pPlayer->v.dmg_inflictor->v.flags & FL_CLIENT)
            {
               // has he a view model ?
               if (strncmp ("models/v_", STRING (pPlayer->v.dmg_inflictor->v.viewmodel), 9) == 0)
               {
                  // build the weapon name
                  sprintf (weapon_name, " with %s", STRING (pPlayer->v.dmg_inflictor->v.viewmodel) + 9);
                  length = strlen (weapon_name); // get length of name
                  for (index = 0; index < length; index++)
                     if (weapon_name[index] == '.')
                        break; // break at the first dot
                  weapon_name[index] = 0; // and terminate the string here
               }
               else
                  strcpy (weapon_name, ""); // else reset the weapon name

               sprintf (ircbot.game_to_irc,
                        "*** %s killed %s%s ***\n",
                        STRING (pPlayer->v.dmg_inflictor->v.netname),
                        STRING (pPlayer->v.netname),
                        weapon_name);
               ircbot.new_game_to_irc = TRUE; // we have something new to send
            }

            // else the killer is another entity
            else
            {
               sprintf (ircbot.game_to_irc,
                        "*** %s killed by %s ***\n",
                        STRING (pPlayer->v.netname),
                        STRING (pPlayer->v.dmg_inflictor->v.classname));
               ircbot.new_game_to_irc = TRUE; // we have something new to send
            }

            playerdeath_sent[player_index] = TRUE; // remember this is done
         }
      }
      else
         playerdeath_sent[player_index] = FALSE; // else don't send anything to IRC
   }

   // do we have some new chat either in IRC or in game ?
   if (ircbot.new_irc_to_game)
   {
      SendIRCToGame (ircbot.irc_to_game); // send the IRC chat to the players' HUDs
      ircbot.new_irc_to_game = FALSE; // clear the flag now we've processed it
   }
   if (ircbot.new_game_to_irc)
   {
      SendGameToIRC (ircbot.game_to_irc); // send the game chat to the IRC channel
      ircbot.new_game_to_irc = FALSE; // clear the flag now we've processed it
   }

   RETURN_META (MRES_IGNORED);
}


void ServerCommand_IRC (void)
{
   // this function gets called when an "irc" server command has been issued from the server
   // console. The arguments are gotten through the same means as with ClientCommand().

   char cmd[256];
   char arg1[256];
   char arg2[256];
   char arg3[256];
   char arg4[256];

   strcpy (cmd, CMD_ARGV (1)); // I hate pointers in these situations
   strcpy (arg1, CMD_ARGV (2));
   strcpy (arg2, CMD_ARGV (3));
   strcpy (arg3, CMD_ARGV (4));
   strcpy (arg4, CMD_ARGV (5));

   // is it a "connect" order ?
   if (strcmp (cmd, "connect") == 0)
   {
      if (arg1[0] != 0)
         strcpy (ircserver.hostname, arg1); // if we have been specified a hostname, set it
      if (atoi (arg2) > 0)
         ircserver.port = atoi (arg2); // if we have been specified a port, set it
      if (arg3[0] != 0)
         strcpy (ircbot.nickname, arg3); // if we have been specified a nickname, set it
      if (arg4[0] != 0)
         strcpy (ircbot.channel, arg4); // if we have been specified a channel, set it

      ServerConsole_printf ("IRC bot: trying to connect to %s:%d as %s and join %s...\n",
                            ircserver.hostname,
                            ircserver.port,
                            ircbot.nickname,
                            ircbot.channel); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_CONNECT; // and tell the IRC bot to connect
   }

   // else is it a "disconnect" order ?
   else if (strcmp (cmd, "disconnect") == 0)
   {
      ServerConsole_printf ("IRC bot: disconnecting from %s.\n", ircserver.hostname); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_DISCONNECT; // tell the IRC bot to disconnect
   }

   // else do we want to set the server hostname ?
   else if ((strcmp (cmd, "server") == 0) && (arg1[0] != 0))
   {
      strcpy (ircserver.hostname, arg1); // if we have been specified a hostname, set it
      ServerConsole_printf ("IRC bot: trying to contact %s...\n", ircserver.hostname); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_DISCONNECT; // and tell the IRC bot to reconnect
   }

   // else do we want to set the server port ?
   else if ((strcmp (cmd, "port") == 0) && (arg1[0] != 0))
   {
      ircserver.port = atoi (arg1); // if we have been specified a port, set it
      ServerConsole_printf ("IRC bot: trying to contact %s on port %d...\n", ircserver.hostname, ircserver.port); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_DISCONNECT; // and tell the IRC bot to reconnect
   }

   // else do we want to set the IRC bot nickname ?
   else if ((strcmp (cmd, "nickname") == 0) && (arg1[0] != 0))
   {
      strcpy (ircbot.nickname, arg1); // if we have been specified a nickname, set it
      ServerConsole_printf ("IRC bot: setting nickname \"%s\"...\n", ircbot.nickname); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_SETNICKNAME; // and tell the IRC bot to change its nickname
   }

   // else do we want to set the IRC channel to join ?
   else if ((strcmp (cmd, "channel") == 0) && (arg1[0] != 0))
   {
      strcpy (ircbot.channel, arg1); // if we have been specified a channel, set it
      ServerConsole_printf ("IRC bot: joining channel %s...\n", ircbot.channel); // tell what we're about to do
      ircbot.order = IRCBOT_ORDER_JOINCHANNEL; // and tell the IRC bot to join this channel
   }

   ircbot.connection_idle_delay = 5; // whatever it was, reset idle delay to 5
   IRCBot_Ping (); // make the bot send a ping so it wakes up when it receives the bounce
   return;
}


void SendIRCToGame (const char *text)
{
   // this function loops through all clients and displays the specified text string in the game
   // chat area of their Head Up Display (HUD). The text is sent as is, without formatting and
   // without colors.

   int client_index;
   edict_t *pClient;

   // if the "SayText" message has not been registered yet...
   if (GET_USER_MSG_ID (PLID, "SayText", NULL) <= 0)
      REG_USER_MSG ("SayText", -1); // register it by the engine

   // now loop through all clients
   for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
   {
      pClient = INDEXENT (client_index); // get a pointer to this client's entity

      if (FNullEnt (pClient) || !(pClient->v.flags & (FL_CLIENT | FL_FAKECLIENT)))
         continue; // if the pointer is invalid or doesn't point to a client, skip this one

      // send the message for display on this client's HUD
      MESSAGE_BEGIN (MSG_ONE_UNRELIABLE, GET_USER_MSG_ID (PLID, "SayText", NULL), NULL, pClient);
      WRITE_BYTE (ENTINDEX (pClient));
      WRITE_STRING (text);
      MESSAGE_END ();
   }

   return; // done
}


void SendGameToIRC (const char *text)
{
   // this function sends the specified game chat text to the IRC channel through the IRC bot

   if (!ircbot.connected_to_server || !ircbot.connected_to_channel)
      return; // reliability check; don't send anything if bot is not connected

   sprintf (ircbot.tx_buffer, "PRIVMSG %s :%s\n", ircbot.channel, ircbot.game_to_irc);
   IRCBot_Talk (); // send the chat string to the IRC server

   return; // done
}


void ServerConsole_printf (char *fmt, ...)
{
   // this function is a convenient varargs to print things to the server console

   va_list argptr;
   static char string[1024];

   va_start (argptr, fmt);
   vsprintf (string, fmt, argptr);
   va_end (argptr);

   SERVER_PRINT (string); // print to console
   return;
}


///////////////////////////////////////////////////////////////////////////////////////////////
// Everything above runs on the main game thread
///////////////////////////////////////////////////////////////////////////////////////////////
// Everything below runs on a different thread
///////////////////////////////////////////////////////////////////////////////////////////////


MT_void StartMultithreadedIRCBot (void *thread_id)
{
   // this function starts a new thread in which the IRC bot will run

   // first off, the thread has successfully started, so give the signal back to the parent
   ircbot.thread_active = TRUE;

   // loop endlessly while we're not asked to quit
   while (ircbot.order != IRCBOT_ORDER_TERMINATETHREAD)
   {
      // has the IRC bot be given an order ?
      if (ircbot.order == IRCBOT_ORDER_CONNECT)
         IRCBot_Connect (); // we have been asked to connect
      else if (ircbot.order == IRCBOT_ORDER_DISCONNECT)
         IRCBot_Disconnect (); // we have been asked to disconnect
      else if (ircbot.order == IRCBOT_ORDER_SETNICKNAME)
         IRCBot_SetNickname (); // we have been asked to change our nickname
      else if (ircbot.order == IRCBOT_ORDER_JOINCHANNEL)
         IRCBot_JoinChannel(); // we have been asked to join the channel

      // else if we are connected...
      else if (ircbot.connected_to_server)
      {
         IRCBot_Listen (); // listen to IRC...
         IRCBot_Think (); // ...and formulate a reply when needed
      }

      // else we are NOT connected yet, or we have been disconnected...
      else
      {
         sleep (ircbot.connection_idle_delay); // so wait for the common idle time
         ircbot.connection_idle_delay = ircbot.connection_idle_delay * 1.2; // increase delay by 20%
         IRCBot_Connect (); // and try to connect by ourselves
      }

      sleep (0); // allow context switching if necessary (thanks botman!)
   }

   // if we reach there, then the thread is about to end
   IRCBot_Disconnect (); // so disconnect from the server if needed
   ircbot.thread_active = FALSE; // give the death signal back to the parent
#ifdef _WIN32
   _endthread (); // suicide (Win32 thread API)
#else
   pthread_exit (NULL); // suicide (POSIX thread API)
#endif
}


void IRCBot_Connect (void)
{
   // this function connects the IRC bot on the IRC server

   int nick_length, irc_nick_length, index;

   // reset the server socket structure
   memset (&ircserver.sockaddr, 0, sizeof (ircserver.sockaddr));

   // get the IRC server's IP address from the host name
   ircserver.hostinfo = gethostbyname (ircserver.hostname);
   if (ircserver.hostinfo == NULL)
   {
      ircbot.error = IRCBOT_ERROR_CANTRESOLVENAME; // set the error level
      return; // break on error
   }

   // fill in the sockaddr server structure with the server hostinfo data
   ircserver.sockaddr.sin_family = AF_INET;
   ircserver.sockaddr.sin_addr.s_addr = inet_addr (inet_ntoa (* (in_addr *) ircserver.hostinfo->h_addr_list[0]));
   ircserver.sockaddr.sin_port = htons (ircserver.port);

   // initialise some variables
   ircbot.connected_to_server = FALSE;
   ircbot.connected_to_channel = FALSE;
   ircbot.irc_socket = NULL;
   ircbot.error = IRCBOT_ERROR_NOERROR;

   // what's this bot's name, again ?
   nick_length = strlen (ircbot.nickname);

   // can we use directly this name as an IRC nickname ?
   if (nick_length < 30)
      irc_nick_length = nick_length; // this nickname fits for an IRC nickname
   else
      irc_nick_length = 30; // IRC nicknames are maximum 30 characters long, looks like

   // build the bot's IRC name
   for (index = 0; index < irc_nick_length; index++)
      if (ircbot.nickname[index] == ' ')
         ircbot.nickname[index] = '_'; // convert spaces to underscores
   ircbot.nickname[irc_nick_length] = 0; // terminate the string

   // create a socket for the IRC bot
   ircbot.irc_socket = socket (AF_INET, SOCK_STREAM, 0);
   if (ircbot.irc_socket == -1)
   {
      ircbot.error = IRCBOT_ERROR_CANTCREATESOCKET; // set the error level
      return; // break on error
   }

   // connect the bot to the IRC server using that socket
   if (connect (ircbot.irc_socket, (sockaddr *) &ircserver.sockaddr, sizeof (ircserver.sockaddr)) == -1)
   {
      ircbot.error = IRCBOT_ERROR_CANTCONNECTTOSERVER; // set the error level
      return; // break on error
   }

   // at this point, the bot has established a socket connection to the IRC server
   ircbot.connected_to_server = TRUE; // so mark it as connected

   // register the user on the IRC server
   sprintf (ircbot.tx_buffer, "USER %s 0 * : \n", ircbot.nickname);
   IRCBot_Talk (); // login step 1: USER

   sprintf (ircbot.tx_buffer, "NICK %s\n", ircbot.nickname);
   IRCBot_Talk (); // login step 2: NICKname

   if (ircbot.order == IRCBOT_ORDER_CONNECT)
      ircbot.order = IRCBOT_ORDER_NOORDER; // we successfully handled the order

   return; // so return
}


void IRCBot_Disconnect (void)
{
   // this function disconnects the IRC bot from the IRC server

   if (!ircbot.connected_to_server)
      return; // reliability check, don't disconnect if the IRC bot is not connected

   // send the QUIT command with the reason for it
   sprintf (ircbot.tx_buffer, "QUIT :Server shutdown\n");
   IRCBot_Talk (); // logout
   closesocket (ircbot.irc_socket); // close the IRC bot socket
   ircbot.connected_to_channel = FALSE; // mark us as no longer member of the IRC channel
   ircbot.connected_to_server = FALSE; // and mark us as disconnected from the IRC server

   if (ircbot.order == IRCBOT_ORDER_DISCONNECT)
      ircbot.order = IRCBOT_ORDER_NOORDER; // we successfully handled the order

   return; // so return
}


void IRCBot_SetNickname (void)
{
   // this function makes the IRC bot change its nickname

   sprintf (ircbot.tx_buffer, "NICK %s\n", ircbot.nickname);
   IRCBot_Talk (); // request to join the specified channel

   if (ircbot.order == IRCBOT_ORDER_SETNICKNAME)
      ircbot.order = IRCBOT_ORDER_NOORDER; // we successfully handled the order

   return; // finished
}


void IRCBot_JoinChannel (void)
{
   // this function makes the IRC bot join a specified channel

   sprintf (ircbot.tx_buffer, "JOIN %s\n", ircbot.channel);
   IRCBot_Talk (); // request to join the specified channel

   if (ircbot.order == IRCBOT_ORDER_JOINCHANNEL)
      ircbot.order = IRCBOT_ORDER_NOORDER; // we successfully handled the order

   return; // finished
}


void IRCBot_Ping (void)
{
   // this function makes the IRC bot send a ping to the IRC server

   send (ircbot.irc_socket, "PING overthere\n", 15, 0); // I have no habit to repeat myself...

   return; // was it worth a function ?
}


void IRCBot_Talk (void)
{
   // this function makes the IRC bot send its transmission buffer to the IRC server

   if (!ircbot.connected_to_server)
      return; // reliability check, don't send anything if the IRC bot is not connected

   // write the transmission buffer to the IRC client socket in direction of the IRC server
   send (ircbot.irc_socket, ircbot.tx_buffer, strlen (ircbot.tx_buffer), 0);
   ircbot.tx_buffer[0] = 0; // reset the tx buffer now that it has been sent

   return; // and return
}


void IRCBot_Listen (void)
{
   // this function makes the IRC bot listen to the data sent by the IRC server and fill in
   // its reception buffer accordingly

   int count = 0;
   static char buffer[256];

   if (!ircbot.connected_to_server)
      return; // reliability check, don't listen to anything if the IRC bot is not connected

   // as long as we don't get a carriage return (13) or we don't reach 256 packets of data...
   while ((buffer[0] != 13) && (count < 256))
   {
      // listen to the socket and stuff the temporary buffer with what we get
      recv (ircbot.irc_socket, buffer, 1, 0);
      strcat (ircbot.rx_buffer, buffer); // concatenate the data in the rx buffer
      count++; // we have processed one packet more

      sleep (0); // allow context switching if necessary (thanks botman!)
   }
   recv (ircbot.irc_socket, buffer, 1, 0); // handle the terminator packet

   return; // and return
}


void IRCBot_Think (void)
{
   // this function examines the reception buffer of the IRC bot and formulate an automatic
   // reply to the IRC server when needed

   static char words[16][256];
   static int index;

   if (!ircbot.connected_to_server)
      return; // reliability check, don't examine anything if the IRC bot is not connected

   if (ircbot.rx_buffer[0] == 0)
      return; // reliability check, don't examine anything if nothing in the reception buffer

   // is the data received typical from an unexpected disconnection ?
   if ((strncmp (ircbot.rx_buffer, "\n\n\n\n\n", 5) == 0)
       || (strncmp (ircbot.rx_buffer, "     ", 5) == 0))
   {
      ircbot.connected_to_channel = FALSE; // bot has been kicked from the IRC channel
      ircbot.connected_to_server = FALSE; // and is no longer connected to the IRC server
      ircbot.rx_buffer[0] = 0; // reset the reception buffer
      return; // don't go any further
   }

   // else we are still connected and we need to examine word by word what the server sends us
   memset (words, 0, sizeof (words));

   // parse the server words and fill our array
   for (index = 0; index < 16; index++)
      strcpy (words[index], GetWord (ircbot.rx_buffer, index)); // store each word

   // is it a "nick already in use" error ?
   if (strcmp ("433", words[1]) == 0)
   {
      strcat (ircbot.nickname, "_"); // append an underscore to his IRC nickname
      IRCBot_SetNickname (); // retry login step 2
   }

   // else is it a PING request ?
   else if (strcmp ("PING", words[0]) == 0)
   {
      sprintf (ircbot.tx_buffer, "PONG %s\n", words[1]);
      IRCBot_Talk (); // send PONG
   }

   // else are we asked for our IRC client version ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp ("VERSION", words[4]) == 0))
   {
      sprintf (ircbot.tx_buffer, "NOTICE %s :%cVERSION %s version %s by %s\n", words[0], 0x01, Plugin_info.name, Plugin_info.version, Plugin_info.author);
      IRCBot_Talk (); // send VERSION name
   }

   // else are we asked for our IRC client ping ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp ("PING", words[4]) == 0))
   {
      sprintf (ircbot.tx_buffer, "NOTICE %s :%cPING %s\n", words[0], 0x01, words[5]);
      IRCBot_Talk (); // send PING
   }

   // else are we asked for our IRC client info ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp ("CLIENTINFO", words[4]) == 0))
   {
      sprintf (ircbot.tx_buffer, "NOTICE %s :%cCLIENTINFO Another metamod plugin by some anonymous weirdo\n", words[0], 0x01);
      IRCBot_Talk (); // send CLIENTINFO
   }

   // else are we asked for our IRC user info ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp ("USERINFO", words[4]) == 0))
   {
      sprintf (ircbot.tx_buffer, "NOTICE %s :%cUSERINFO Mind your own business!\n", words[0], 0x01);
      IRCBot_Talk (); // send CLIENTINFO
   }

   // else are we asked for our IRC client local time ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp ("TIME", words[4]) == 0))
   {
      sprintf (ircbot.tx_buffer, "NOTICE %s :%cTIME Uptime on map %s: %d min %d seconds\n", words[0], 0x01, STRING (gpGlobals->mapname), (int) gpGlobals->time / 60, (int) gpGlobals->time % 60);
      IRCBot_Talk (); // send CLIENTINFO
   }

   // else is it a MOTD notice ?
   else if ((strcmp ("376", words[1]) == 0) || (strcmp ("422", words[1]) == 0))
      IRCBot_JoinChannel (); // it's time to join our channel

   // else is it a NAMES list ? (such a list is sent when we have joined a channel)
   else if (strcmp ("366", words[1]) == 0)
      ircbot.connected_to_channel = TRUE; // bot is now a member of the channel

   // else is it a notification that we have been kicked out of the channel ?
   else if (strcmp ("KICK", words[2]) == 0)
   {
      ircbot.connected_to_channel = FALSE; // bot is no more a member of the channel
      IRCBot_JoinChannel (); // try again to enter
   }

   // else is it someone's chat in the main channel ?
   else if ((strcmp ("PRIVMSG", words[2]) == 0) && (strcmp (ircbot.channel, words[3]) == 0))
   {
      // build the string to be displayed on the players' HUDs
      sprintf (ircbot.irc_to_game, "%s :    %s", words[0], &ircbot.rx_buffer[GetNextWordStartOffset (ircbot.rx_buffer, 3)]);

      // remove trailing carriage returns and line feeds from the IRC server
      while (((index = strlen (ircbot.irc_to_game) - 1) >= 0)
             && ((ircbot.irc_to_game[index] == 0x0D) || (ircbot.irc_to_game[index] == 0x0A)))
         ircbot.irc_to_game[index] = 0; // shorten the string

      strcat (ircbot.irc_to_game, "\n"); // terminate the string with a standard line feed
      ircbot.new_irc_to_game = TRUE; // notify the main thread that there is a new IRC chat message
   }

   // that's all we can do, so reset the reception buffer
   ircbot.rx_buffer[0] = 0;
   return; // and return
}


const char *GetWord (const char *string, int word_number)
{
   // this function returns the wanted word in a string

   static char word_returned[256];
   int length, i, index = 0, word_count = 0, fieldstart, fieldstop;

   word_returned[0] = 0; // clear the character string that will hold the word to be returned
   length = strlen (string); // get the length of the original string

   // while we have not reached the end of the string
   while ((index < length) && (word_count <= word_number))
   {
      while ((index < length) && ((string[index] == ' ')
                                  || (string[index] == ':')
                                  || (string[index] == '!')
                                  || (string[index] == '~')
                                  || (string[index] == 0x01)
                                  || (string[index] == 0x0D)
                                  || (string[index] == 0x0A)))
         index++; // ignore spaces, smileys, separators, carriage returns and end of lines

      fieldstart = index; // save field start position
      while ((index < length) && (string[index] != ' ') 
                              && (string[index] != ':')
                              && (string[index] != '!')
                              && (string[index] != '~')
                              && (string[index] != 0x01)
                              && (string[index] != 0x0D)
                              && (string[index] != 0x0A))
         index++; // reach end of field
      fieldstop = index - 1; // save field stop position

      // is this word we just processed the wanted one ?
      if (word_count == word_number)
      {
         for (i = fieldstart; i <= fieldstop; i++)
            word_returned[i - fieldstart] = string[i]; // store the field value in a string
         word_returned[i - fieldstart] = 0; // terminate the string
         break; // and exit the while loop
      }

      word_count++; // we have processed one word more
   }

   return (&word_returned[0]); // return a pointer to the wanted word
}


int GetNextWordStartOffset (const char *string, int word_number)
{
   // this function returns the position in the character string of the beginning of the word
   // that is immediately after the wanted word

   int length, index = 0, word_count = 0, fieldstart, fieldstop;

   length = strlen (string); // get length of string

   // while we have not reached the end of the string
   while ((index < length) && (word_count <= word_number))
   {
      while ((index < length) && ((string[index] == ' ')
                                  || (string[index] == ':')
                                  || (string[index] == '!')
                                  || (string[index] == '~')
                                  || (string[index] == 0x01)
                                  || (string[index] == 0x0D)
                                  || (string[index] == 0x0A)))
         index++; // ignore spaces, smileys, separators, carriage returns and end of lines

      fieldstart = index; // save field start position
      while ((index < length) && (string[index] != ' ') 
                              && (string[index] != ':')
                              && (string[index] != '!')
                              && (string[index] != '~')
                              && (string[index] != 0x01)
                              && (string[index] != 0x0D)
                              && (string[index] != 0x0A))
         index++; // reach end of field
      fieldstop = index - 1; // save field stop position

      // is this word we just processed the wanted one ?
      if (word_count == word_number)
      {
         while ((index < length) && ((string[index] == ' ') || (string[index] == ':')))
            index++; // skip the spaces and the ':' after the word
         return (index); // return the offset at which the next word begins
      }

      word_count++; // we have processed one word more
   }

   return (strlen (string)); // word not found, return the \0 index at the end of the string
}
